import unittest

import withgui

class TestImplementation(dict):
    def set_property(self, name, value):
        self[name] = value
    def get_property(self, name):
        return self[name]

class TestSpec(unittest.TestCase):

    def test_descendents(self):
        gui = withgui.Column()
        with gui.row():
            gui.label()
        self.assertEquals(len(gui.children), 1)
        self.assertEquals(len(gui.children[0].children), 1)

    def test_form_row_label(self):
        self._test_form_row('label')
    def test_form_row_widet(self):
        self._test_form_row('text')
    def _test_form_row(self, thing):
        with withgui.Form() as gui:
            getattr(gui, thing)()
            getattr(gui, thing)()
            gui.submit('ok')
        self.assertEquals(len(gui.children), 2)
        self.assertEquals(len(gui.rows), 2)
        self.assertEquals(len(gui.buttons), 1)

    def test_form_row_full(self):
        with withgui.Form() as gui:
            gui.label(); gui.text(); gui.help()
            gui.help(); gui.text(); gui.label()
            gui.help(); gui.label(); gui.text()
        self.assertEquals(len(gui.children), 9)
        self.assertEquals(len(gui.rows), 3)

    def test_form_row_helper(self):
        with withgui.Form() as gui:
            r1 = gui.row('label1', gui.selection([]), 'help1')
            r2 = gui.row('label2', gui.text(), 'help2')
        self._assert_row_correct(gui, r1, r2)

    def test_form_row_mixed(self):
        with withgui.Form() as gui:
            r1 = gui.row('label1', gui.selection([]), 'help1')
            gui.label('label2')
            r2 = gui.text()
            gui.help('help2')
        self._assert_row_correct(gui, r1, r2)

    def _assert_row_correct(self, gui, r1, r2):
        self.assertEquals(len(gui.children), 6)
        self.assertEquals(len(gui.rows), 2)
        self.assertEquals(gui.rows[0].label.value, 'label1')
        self.assert_(isinstance(r1, withgui.Selection))
        self.assertEquals(gui.rows[0].help.value, 'help1')
        self.assertEquals(gui.rows[1].label.value, 'label2')
        self.assert_(isinstance(r2, withgui.Text))
        self.assertEquals(gui.rows[1].help.value, 'help2')

    def test_form_submit_cancel(self):
        with withgui.Form() as gui:
            gui.text()
            with gui.submit('Go!') as go:
                def on_click(): pass
            with gui.cancel('No stop!') as stop:
                def on_click(): pass
        self.assertEquals(len(gui.buttons), 2)
        self.assert_('on_click' in go.event_handlers)
        self.assert_('on_click' in stop.event_handlers)
        self.assertNotEqual(go.event_handlers['on_click'],
            stop.event_handlers['on_click'])

    def test_local_ignored(self):
        with withgui.Row() as gui:
            name = gui.label
        self.assert_('name' not in gui.named_widgets)

    def test_fixed_args(self):
        l = withgui.Label('Name')
        self.assertEquals(l.value, 'Name')

    def test_kw_args(self):
        l = withgui.Label(value='Name')
        self.assertEquals(l.value, 'Name')

    def test_reject_unknown_kw(self):
        self.assertRaises(KeyError, withgui.Label, bogus='Name')

    def test_decorator(self):
        b = withgui.Button()
        @b.on_click
        def foo(*args): pass
        self.assert_('on_click' in b.event_handlers)

    def test_events_with(self):
        with withgui.Form() as gui:
            with gui.button():
                def on_click():
                    pass
        button = gui.children[0]
        self.assert_('on_click' in button.event_handlers)
        self.assert_('on_click' not in gui.event_handlers)

    def test_form_with_name(self):
        with withgui.Form() as gui:
            foo = gui.text()
            bar = gui.text()
        self.assert_(gui['foo'])
        self.assert_(gui['bar'])

    def test_access(self):
        with withgui.Column() as gui:
            with gui.form() as form:
                foo = gui.text()
                gui.text()
        self.assert_(gui['form']['foo'])
        self.assert_(gui['form']['.text'])

    def test_set_property(self):
        l = withgui.Label()
        l.value = 1
        self.assertEquals(l.value, 1)
        self.assertEquals(l.settings['value'], 1)

    def test_set_property_implementation(self):
        l = withgui.Label()
        l.foreground = withgui.red
        self.assertEquals(l.foreground, withgui.red)
        self.assertEquals(l.settings['foreground'], withgui.red)
        l.implementation = TestImplementation()
        l.foreground = withgui.blue
        self.assertEquals(l.settings['foreground'], withgui.red)
        self.assertEquals(l.foreground, withgui.blue)
        self.assertEquals(l.implementation.get_property('foreground'), withgui.blue)

    def test_colour_prop_tuple(self):
        self.assertEquals(withgui.prop_color((.1, .2, .3, .4)),
            withgui.Color(.1, .2, .3, .4))
    def test_colour_prop_name(self):
        self.assertEquals(withgui.prop_color('white'),
            withgui.Color(1, 1, 1, 1))
    def test_colour_prop_hex(self):
        self.assertEquals(withgui.prop_color('FFF'),
            withgui.Color(1, 1, 1, 1))
        self.assertEquals(withgui.prop_color('FFF3'),
            withgui.Color(1, 1, 1, .2))
        self.assertEquals(withgui.prop_color('FFFFFF'),
            withgui.Color(1, 1, 1, 1))
        self.assertEquals(withgui.prop_color('FFFFFF33'),
            withgui.Color(1, 1, 1, .2))
        self.assertEquals(withgui.prop_color('3F0'),
            withgui.Color(.2, 1, 0, 1))
        self.assertEquals(withgui.prop_color('33FF00'),
            withgui.Color(.2, 1, 0, 1))

    def test_color_or_image_prop(self):
        self.assertEquals(withgui.prop_color_or_image('foo.jpg'),
            'foo.jpg')
        self.assertEquals(withgui.prop_color_or_image('red'),
            withgui.Color(1, 0, 0, 1))
        self.assertEquals(withgui.prop_color_or_image('http://foo.example/foo.jpg'),
            'http://foo.example/foo.jpg')

    def test_animation(self):
        with withgui.Frame() as gui:
            with gui.label('hello'):
                def animate(label):
                    dt = yield
        label = gui.children[0]
        self.assert_(label.animate)
        self.assert_(not gui.animate)

    def test_run_for(self):
        l = []
        @withgui.run_for(2)
        def animate(widget):
            while 1:
                s = yield
                l.append(s)

        g = animate(None)
        g.send(None)
        g.send(1)
        self.assertEquals(l, [.5])
        self.assertRaises(StopIteration, g.send, 1)
        self.assertEquals(l, [.5, 1])

    def test_gui_after(self):
        'check that an after spec works on the gui'
        with withgui.Frame() as gui:
            @gui.after(2)
            def foo(): pass
        self.assertEquals(gui.after_actions, [(2, foo)])


    def test_child_after(self):
        'check that an after spec works on a child widget'
        with withgui.Frame() as gui:
            label = gui.label('')
            @label.after(2)
            def foo(): pass

