import re
import os
import sys
import inspect
import functools

# a whole lotta constants
center = 'center'
left = 'left'
right = 'right'
x = 'x'
y = 'y'
both = 'both'
n = 'n'
e = 'e'
s = 's'
w = 'w'
nw = 'nw'
ne = 'ne'
sw = 'sw'
se = 'se'

# color code below originally from pyglet contrib/layout (with extensions)
class Color(tuple):
    def __new__(cls, r, g, b, a=1):
        return tuple.__new__(cls, (r, g, b, a))

    def to_hex(self, a=False):
        if a:
            return '%02x%02x%02x%02x'%(self[0]*255, self[1]*255,
                self[2]*255, self[3]*255)
        else:
            return '%02x%02x%02x'%(self[0]*255, self[1]*255, self[2]*255)

    @classmethod
    def from_hex(cls, hex):
        if len(hex) == 3:
            return Color(int(hex[0], 16) / 15., 
                         int(hex[1], 16) / 15.,
                         int(hex[2], 16) / 15.)
        elif len(hex) == 4:
            return Color(int(hex[0], 16) / 15., 
                         int(hex[1], 16) / 15.,
                         int(hex[2], 16) / 15.,
                         int(hex[3], 16) / 15.)
        elif len(hex) == 6:
            return Color(int(hex[0:2], 16) / 255.,
                         int(hex[2:4], 16) / 255.,
                         int(hex[4:6], 16) / 255.)
        else:
            return Color(int(hex[0:2], 16) / 255.,
                         int(hex[2:4], 16) / 255.,
                         int(hex[4:6], 16) / 255.,
                         int(hex[6:8], 16) / 255.)

# common color names
Color.names = {
    'maroon':   Color.from_hex('800000'),
    'red':      Color.from_hex('ff0000'),
    'orange':   Color.from_hex('ffa500'),
    'yellow':   Color.from_hex('ffff00'),
    'olive':    Color.from_hex('808000'),
    'purple':   Color.from_hex('800080'),
    'fuschia':  Color.from_hex('ff00ff'),
    'white':    Color.from_hex('ffffff'),
    'lime':     Color.from_hex('00ff00'),
    'green':    Color.from_hex('008000'),
    'navy':     Color.from_hex('000080'),
    'blue':     Color.from_hex('0000ff'),
    'aqua':     Color.from_hex('00ffff'),
    'teal':     Color.from_hex('008080'),
    'black':    Color.from_hex('000000'),
    'silver':   Color.from_hex('c0c0c0'),
    'gray':     Color.from_hex('808080'),
}

# convenience
maroon = Color.names['maroon']
red = Color.names['red']
orange = Color.names['orange']
yellow = Color.names['yellow']
olive = Color.names['olive']
purple = Color.names['purple']
fuschia = Color.names['fuschia']
white = Color.names['white']
lime = Color.names['lime']
green = Color.names['green']
navy = Color.names['navy']
blue = Color.names['blue']
aqua = Color.names['aqua']
teal = Color.names['teal']
black = Color.names['black']
silver = Color.names['silver']
gray = Color.names['gray']

def prop_color(value):
    '''Parse a color value which is one of:

    name    a color name (CSS 2.1 standard colors)
    RGB     a three-value hex color
    RRGGBB  a three-value hex color
    '''
    if not isinstance(value, str):
        return value
    if value in Color.names:
        return Color.names[value]
    return Color.from_hex(value)
def prop_color_or_image(value):
    if not isinstance(value, str):
        return value
    if value in Color.names:
        return Color.names[value]
    if re.match(r'[0-9a-f]{3-8}', value, re.I):
        return Color.from_hex(value)
    # image filename
    return value
def prop_image(value):
    # TODO check?
    return value
def prop_direction(value):
    # TODO check
    return value
def prop_justify_options(value):
    assert value in ('left', 'right', 'center')
    return value
def prop_fill(value):
    assert value in ('x', 'y', 'both')
    return value
def prop_boolean(value):
    return bool(value)
def prop_number(value):
    # TODO check
    return value
def prop_width(value):
    # TODO check
    return value
def prop_height(value):
    # TODO check
    return value
def prop_passthrough(value): return value

class PropertiesDict(dict):
    '''Extension of dict that limits the keys to those in the properties
    dict and enforces value types (and handles some type conversions).
    '''
    def __init__(self, properties, d):
        self.properties = properties
        self.update(d)
    def __setitem__(self, key, value):
        if key not in self.properties:
            raise KeyError('%s is not a valid property'%key)
        value = self.properties[key](value)
        super(PropertiesDict, self).__setitem__(key, value)
    def update(self, d):
        for key in d:
            if key not in self.properties:
                raise KeyError('%s is not a valid property'%key)
            d[key] = self.properties[key](d[key])
        super(PropertiesDict, self).update(d)

class WidgetBase(object):
    # widget registry
    widget_classes = {}
    @classmethod
    def register_class(cls, klass, name=None):
        if name is None:
            name = klass.__name__.lower()
        cls.widget_classes[name] = klass

    # instance attribute defaults
    _gui = None
    implementation = None
    animate = None
    name = None         # this will be filled in by the locals() magic
    parent = None

    # property specs
    properties = {}
    defaults = {}
    fixed_args = []

    def __init__(self, *args, **kw):
        if 'parent' in kw:
            # I'm a child widget
            self.parent = kw.pop('parent')
            self.gui = self.parent.gui
            self.parent.add_child(self)
        else:
            # I'm the top-level widget so act like one
            self.stack = [self]                 # CIRCULAR REF here
            self.gui = self                     # CIRCULAR REF here
            self.after_actions = []
            self.window_settings = dict(title='withgui')
            self.resource = Resource()

        self.children = []
        self.event_handlers = {}
        self.named_widgets = {}

        # TODO allow layouts to accept a list of child widgets
        # TODO rename "parent" to "container" or "box" or something

        settings = dict(self.defaults)

        if args:
            kw.update(dict(zip(self.fixed_args, args)))

        # callbacks / animate
        for k in dict(kw):
            if k == 'animate' or k.startswith('on_'):
                self.add_handler(k, kw.pop(k))

        settings.update(kw)
        self.settings = PropertiesDict(self.properties, settings)

        # if we're already physical then add the widget to the actual gui
        if self.parent and self.parent.implementation:
            # TODO handle no parent - create new top-level window
            self.parent.implementation.add_child(self)

    def __getattr__(self, name):
        if name in self.widget_classes:
            if self is self.gui:
                parent = self.gui.stack[-1]
            else:
                parent = self
            return functools.partial(self.widget_classes[name],
                parent=parent)

        if self.implementation is not None:
            return self.implementation.get_property(name)
        elif name in self.properties:
            return self.settings[name]
        try:
            return self.__dict__[name]
        except KeyError:
            raise AttributeError('%r has no attribute %s'%(self, name))

    def __setattr__(self, name, value):
        if name in ('parent', 'implementation', 'gui', 'extra'):
            self.__dict__[name] = value
        elif self.implementation is not None:
            self.implementation.set_property(name, value)
        elif name in self.properties:
            self.settings[name] = self.properties[name](value)
        else:
            self.__dict__[name] = value


    def __call__(self, func):
        '''Decorate some function - or rather schlurp it into this widget
        spec as some event handler or animator.
        '''
        # TODO refactor so this isn't duplicated in __exit__
        # are we invoked as a function decorator?
        if not hasattr(func, '__call__'):
            raise ValueError('only callable as a decorator')
        name = func.func_name
        self.add_handler(name, func)
        # this for the locals() hax
        func.name = name

    def add_handler(self, name, callable):
        if name == 'animate':
            # hi, I'm a generator, so generate me
            self.animate = callable(self)
            self.animate.send(None)
        elif name.startswith('on_'):
            # TODO validate name
            self.event_handlers[name] = callable
        else:
            # not a name I recognise - ignore it
            return


    # overridable child adder (for example see Form.add_child)
    def add_child(self, child):
        self.children.append(child)
        child.parent = self

    def remove_child(self, child):
        self.children.remove(child)
        child.parent = None

    def __getitem__(self, item):
        if isinstance(item, str):
            if item[0] == '.':
                name = item[1:].lower()
                return [child for child in self.children
                    if child.__class__.__name__.lower() == name]
            return self.named_widgets[item]
        else:
            return self.children[item]

    def __enter__(self):
        # save off the current set of names in the locals()
        f = sys._getframe(1)
        locals = inspect.getargvalues(f)[3]
        self._names = set((k, id(v)) for k, v in locals.items())
        self.gui.stack.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # determine whether any new local variables were created during the
        # context manager's lifespan and if so (and they're not already
        # claimed) then add to this object's named objects / event handlers
        f = sys._getframe(1)
        locals = inspect.getargvalues(f)[3]

        # TODO don't store locals names on the object, be smarter about
        # with "scopes" and store a new stack or something
        # TODO this is currently buggy - see the second on_click in
        # form_simple.py

        for name in locals:
            o = locals[name]
            key = (name, id(o))

            if key in self._names:
                continue

            if getattr(o, 'name', None) is not None:
                continue

            if o is self:
                # this is my with statement, don't name me!
                continue

            if isinstance(o, WidgetBase):
                self.named_widgets[name] = o

            elif hasattr(o, '__call__'):
                if name == 'animate':
                    # TODO check signature for dt?
                    # hi, I'm a generator, so generate and kick me off
                    self.animate = o(self)
                    self.animate.send(None)
                elif name.startswith('on_'):
                    self.event_handlers[name] = o
                else:
                    # not a name I recognise - ignore ir
                    continue

            else:
                # not a widget or callable, ignore it
                continue

            o.name = name
        self.gui.stack.pop()

    def dump(self, level=''):
        print '%s%s %r'%(level, self.__class__.__name__.lower(),
            self.settings)
        for c in self.children:
            c.dump(level + '  ')

    def after(self, delay, action=None):
        '''Schedule some delayed action.

        For example::

            @gui.after(2)
            def action():
                ... do something

        is equivalent to::

            gui.after(2, action)
        '''
        if action is not None:
            self.gui.after_actions.append((delay, action))
            return action
        def f(action):
            self.gui.after_actions.append((delay, action))
            return action
        return f

    def run(self):
        #from withsimplui import GUI
        #from withqt4 import GUI
        from withtk import GUI
        #from withkytten import GUI

        # the assignment here isn't strictly necessary
        self.gui = GUI(self)
        self.gui.run()

    def stop(self, data=None):
        return self.gui.stop(data)

    def window(self, **kw):
        # TODO validate the keywords / values
        self.gui.window_settings.update(kw)

    def destroy(self):
        # remove me from my parent and if I'm implemented then destroy that
        # too
        self.parent.remove_child(self)
        if self.implementation:
            self.implementation.destroy()

# TODO Slider
# TODO Checkbox
# TODO Color
# TODO Marked-up text
class Frame(WidgetBase):
    # TODO other properties
    properties = dict(
        padding=prop_number,
        background=prop_color_or_image,
        width=prop_width,
        height=prop_height,
    )
WidgetBase.register_class(Frame)

class Label(WidgetBase):
    fixed_args = ('value',)
    defaults = dict(value='', anchor=nw)
    properties = dict(
        foreground=prop_color,
        background=prop_color,
        anchor=prop_direction,
        expand=prop_boolean,
        justify=prop_justify_options,
        margin=prop_number,
        fill=prop_fill,
        x=prop_number,
        y=prop_number,
        value=prop_passthrough,
    )
WidgetBase.register_class(Label)

class Image(WidgetBase):
    fixed_args = ('value',)
    defaults = dict(value=None, anchor=nw)
    properties = dict(
        anchor=prop_direction,
        justify=prop_justify_options,
        margin=prop_number,
        x=prop_number,
        y=prop_number,
        value=prop_image,
    )
WidgetBase.register_class(Image)

class Canvas(WidgetBase):
    # TODO other properties?
    properties = dict(
        padding=prop_number,
        width=prop_width,
        height=prop_height,
        background=prop_color_or_image,
    )
    widget_classes = dict(
        label = Label,
        image = Image,
    )
WidgetBase.register_class(Canvas)

class Column(Frame):
    pass
WidgetBase.register_class(Column)

class Row(Frame):
    pass
WidgetBase.register_class(Row)

class FormRow(object):
    '''This encapsulates a label, widget and help text for a single row
    (value) in a Form.
    '''
    label = None
    widget = None
    help = None
    def __init__(self, gui=None):
        self.gui = gui
    def needs(self, widget):
        if isinstance(widget, Label):
            return self.label is None
        elif isinstance(widget, Help):
            return self.help is None
        return self.widget is None
    def add(self, widget):
        if isinstance(widget, Label):
            self.label = widget
        elif isinstance(widget, Help):
            self.help = widget
        else:
            self.widget = widget

class Form(WidgetBase):
    def __init__(self, *args, **kw):
        super(Form, self).__init__(*args, **kw)
        self.rows = [FormRow(self.gui)]
        self.buttons = []

    def add_child(self, child):
        if isinstance(child, FormRow):
            self.rows.append(child)
        elif isinstance(child, FormButton):
            self.buttons.append(child)
        else:
            self.children.append(child)
            child.parent = self
            if not self.rows[-1].needs(child):
                self.rows.append(FormRow())
            self.rows[-1].add(child)

    def row(self, label=None, widget=None, help=None):
        '''Create a row in one go. The label and help are passed in as
        strings. The widget is separately constructed.

        Returns the widget for convenience (may be bound to a local name
        for later access or manipulation).
        '''
        if label is not None:
            assert not isinstance(label, Label)
            Label(label, parent=self)
        if widget is not None:
            assert isinstance(widget, WidgetBase)
        if help is not None:
            assert not isinstance(help, Help)
            Help(help, parent=self)
        # for convenience this will allow local binding of the widget to a
        # name
        return widget

WidgetBase.register_class(Form)

class Text(WidgetBase):
    fixed_args = ('value',)
    properties = dict(
        value=prop_passthrough,
    )
WidgetBase.register_class(Text)

class Selection(WidgetBase):
    fixed_args = ('options', 'value')
    properties = dict(
        options=prop_passthrough,
        value=prop_passthrough,
    )
WidgetBase.register_class(Selection)

class Help(WidgetBase):
    fixed_args = ('value',)
    properties = dict(
        value=prop_passthrough,
    )
WidgetBase.register_class(Help)

class Button(WidgetBase):
    fixed_args = ('value',)
    properties = dict(
        foreground=prop_color,
        background=prop_color,
        anchor=prop_direction,
        justify=prop_justify_options,
        margin=prop_number,
        fill=prop_fill,
        x=prop_number,
        y=prop_number,
        value=prop_passthrough,
    )
    def on_click(self, handler):
        self.event_handlers['on_click'] = handler
WidgetBase.register_class(Button)

class FormButton(Button):
    # group submit / cancel button classes
    pass
class Submit(FormButton):
    pass
WidgetBase.register_class(Submit)

class Cancel(FormButton):
    pass
WidgetBase.register_class(Cancel)

class Resource(object):
    def __init__(self):
        self.paths = [os.path.abspath(os.path.dirname(sys.argv[0]))]

    def add(self, path):
        # TODO allow ZIP files etc.
        self.paths.append(path)

    def lookup(self, name):
        for path in self.paths:
            p = os.path.join(path, name)
            if os.path.isfile(os.path.join(path, name)):
                return p
    
def run_for(seconds):
    '''Become a generator over the other animation generator f() which
    expects to be passed values from 0 to 1 over the specified time in
    seconds.
    '''
    def g(f):
        def animate(w, seconds=seconds, f=f):
            t = 0.0
            f = f(w)
            f.send(None)
            while t < seconds:
                dt = yield
                t += dt
                if t > seconds:
                    f.send(1)
                else:
                    f.send(t / seconds)
        return animate
    return g

