import sys
import time
import urllib
import functools

# py3k compat
try:
    from io import StringIO
except ImportError:
    from StringIO import StringIO
try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk

from PIL import Image as PIL_Image
from PIL import ImageTk

import withgui

class PackLayout(object):
    pack_side = tk.TOP
    def layout(self, child):
        # TODO check child DOES NOT have x, y and raise sensible error
        pack = dict(side=self.pack_side)
        if 'fill' in child.args:
            pack['fill'] = child.args['fill']
        if 'anchor' in child.args:
            pack['anchor'] = child.args['anchor']
        if 'expand' in child.args:
            pack['expand'] = child.args['expand']
        child.tk_widget.pack(**pack)

class GUI(PackLayout):
    '''The owner of the implementaton of a withgui spec.

    Holds the toplevel widget Implementation, the animators and the
    after_actions.
    '''
    def __init__(self, spec):
        self.animators = []

        self.tk_widget = tk.Tk()

        # TODO more window properties
        self.tk_widget.title(spec.window_settings['title'])
        #self.tk_widget.minsize(200, 100)
        self.tk_widget.geometry('+100+100')

        # add space at the bottom to account for OS X's resize handle
        if sys.platform == 'darwin':
            status = tk.Label(self.tk_widget, text="")
            status.pack(side=tk.BOTTOM, fill=tk.X)

        # construct the spec
        self.gui = self
        self.toplevel = Implementation.construct(spec, self)

        # add delayed actions
        for delay, action in spec.after_actions:
            self.tk_widget.after(int(delay*1000), action)

        # set up top-level event handlers
        for name, action in spec.event_handlers.items():
            if name == 'on_mouse':
                # TODO all buttons
                toplevel.bind("<Button-1>", MouseEvent(spec, action, left=True))

    def update(self):
        dt = time.time() - self.ts
        self.ts = time.time()
        for animator in list(self.animators):
            try:
                animator.send(dt)
            except StopIteration:
                self.animators.remove(animator)
        # TODO handle timing skew
        # TODO only schedule this if there's active animators
        self.tk_widget.after(int(1000/20), self.update)

    def run(self):
        self.ts = time.time()
        # TODO only schedule this if there's active animators
        self.tk_widget.after(int(1000/20), self.update)
        self.tk_widget.focus()
        self.tk_widget.lift()
        self.tk_widget.mainloop()

    def stop(self, data=None):
        self.tk_widget.destroy()
        return data

class HasProperties(object):
    # mapping withgui properties to Tkinter properties
    prop_int = lambda self, v: int(v)
    # TODO: check ALL uses of this for better type
    prop_passthrough = lambda self, v: v
    prop_color = lambda self, v: '#'+v.to_hex()
    prop_fill = lambda self, v: dict(x=tk.X, y=tk.Y, both=tk.BOTH)[v]
    def prop_image(self, spec):
        # TODO accept more URLs
        if spec.startswith('http://'):
            s = StringIO(urllib.urlopen(spec).read())
            return ImageTk.PhotoImage(PIL_Image.open(s))

        # find the image
        path = self.spec.gui.resource.lookup(spec)
        if path is None:
            raise ValueError("can't locate image file %s"%spec)
        return ImageTk.PhotoImage(PIL_Image.open(path))
    def prop_color_or_image(self, spec):
        # can't isinstance() here because __main__.Color != withgui.Color!
        if hasattr(spec, 'to_hex'):
            return self.prop_color(spec)
        return self.prop_image(spec)
    # TODO more anchors
    prop_anchor = lambda self, v: dict(center=tk.CENTER, sw=tk.SW,
        nw=tk.NW).get(v)

    # properties are names mapped to the Tkinter name and conversion callable
    base_properties = dict(
        justify=('justify', prop_passthrough),
        x=('x', prop_int),
        y=('y', prop_int),
        width=('width', prop_passthrough),
        height=('height', prop_passthrough),
        foreground=('foreground', prop_color),
        background=('background', prop_color),

        # packing args
        fill=('fill', prop_fill),
        anchor=('anchor', prop_anchor),
        expand=('expand', prop_int),

        value=('value', prop_passthrough),
    )
    properties = {}

    def generate_properties(self, spec):
        if spec.animate:
            self.gui.animators.append(spec.animate)
        d = {}
        for k,v in spec.settings.items():
            if k in self.properties:
                k, p = self.properties[k]
            else:
                k, p = self.base_properties[k]
            d[k] = p(self, v)
        return d

    def translate_value(self, name, value):
        if name in self.properties:
            k, p = self.properties[name]
        else:
            k, p = self.base_properties[name]
        return k, p(self, value)

class Implementation(HasProperties):
    '''Wrapper around a Tkinter widget and optionally a separate Tkinter
    variable allowing read/write access to the widget's properties
    including its current value.
    '''
    implementation_classes = {}
    @classmethod
    def register_class(cls, klass, name=None):
        if name is None:
            name = klass.__name__
        cls.implementation_classes[name] = klass

    @classmethod
    def construct(cls, spec, parent):
        klass = cls.implementation_classes[spec.__class__.__name__]
        return klass(spec, parent)

    def __init__(self, spec, parent):
        self.spec = spec
        self.parent = parent
        self.gui = parent.gui
        self.tk_widget = None
        self.args = self.generate_properties(spec)

        self.create_widget()

        parent.layout(self)

        # set up event handlers
        if 'on_mouse' in spec.event_handlers:
            # TODO all buttons
            image.bind("<Button-1>", MouseEvent(spec,
                spec.event_handlers['on_mouse'], left=True))

        self.spec.implementation = self

    def create_widget(self):
        '''Create the underlying tk_widget / tk_variable.
        '''
        raise NotImplementeError('create_widget must be implemented')

    def set_value(self, value):
        self.tk_variable.set(value)
    def get_value(self):
        return self.tk_variable.get()
    def set_property(self, name, value):
        name, value = self.translate_value(name, value)
        if name == 'value':
            self.set_value(value)
            return
        if name == 'x':
            self.tk_widget.x = value
            self.tk_widget.place(x=value)
        elif name == 'y':
            self.tk_widget.y = value
            self.tk_widget.place(y=value)
        else:
            self.tk_widget.config(**{name: value})
    def get_property(self, name):
        if name == 'value':
            return self.get_value()
        elif name == 'x':
            return self.tk_widget.x
        elif name == 'y':
            return self.tk_widget.y
        else:
            return getattr(self.tk_widget, name)

    def destroy(self):
        self.spec.implementation = None
        self.parent = None
        self.gui = None
        self.args = None
        self.tk_widget.unbind("<Button-1>")
        self.tk_widget.destroy()
        self.tk_widget = self.tk_variable = None

class Frame(Implementation):
    properties = dict(
        side=('side', HasProperties.prop_passthrough),
        background=('background', HasProperties.prop_color_or_image),
    )
    conprops = 'width height side background'.split()

    def create_widget(self):
        # oh, for dict comprehensions
        kw = {}
        for name in self.conprops:
            if name in self.args:
                kw[name] = self.args[name]

        image = self.args.get('image')
        if image is not None:
            if 'width' not in kw:
                kw['width'] = image.width()
            if 'height' not in kw:
                kw['height'] = image.height()

        self.tk_widget = tk.Frame(self.parent.tk_widget, **kw)

        if image is not None:
            self._image = image
            self.image_widget = tk.Label(self.tk_widget,
                image=self._image)
            self.image_widget.place(x=0, y=0, anchor=tk.NW)

        # create children
        # this is is done here so it may be customised per widget (well,
        # frame vs. canvas)
        for child in self.spec.children:
            Implementation.construct(child, self)

    def add_child(self, spec):
        Implementation.construct(spec, self)

    def layout(self, child):
        '''Add children by x, y coordinates.
        '''
        # TODO check child has x, y and raise sensible error
        x = child.args['x']
        y = child.args['y']
        anchor = child.args.get('anchor', tk.CENTER)
        child.tk_widget.place(x=x, y=y, anchor=anchor)
        child.x = x
        child.y = y
        

Implementation.register_class(Frame)

class Column(PackLayout, Frame):
    pass
Implementation.register_class(Column)

class Row(PackLayout, Frame):
    pack_side = tk.LEFT
Implementation.register_class(Row)

class Label(Implementation):
    properties = dict(
        value=('text', HasProperties.prop_passthrough),
    )
    # TODO better name
    conprops = 'background foreground'.split()

    def create_widget(self):
        # oh, for dict comprehensions
        kw = {}
        for name in self.conprops:
            if name in self.args:
                kw[name] = self.args[name]

        kw['textvariable'] = self.tk_variable = tk.StringVar()
        self.tk_variable.set(self.args.get('text', ''))
        self.tk_widget = tk.Label(self.parent.tk_widget, **kw)

Implementation.register_class(Label)

class Help(Label):
    def __init__(self, spec, parent, pack={}):
        return self.add_Label(spec, parent, pack=pack)
Implementation.register_class(Help)

class Image(Implementation):
    properties = dict(
        value=('image', HasProperties.prop_image),
    )

    def create_widget(self):
        self.tk_widget = tk.Label(self.parent.tk_widget,
            image=self.args['image'])

    def set_value(self, value):
        self.tk_widget.config(image=value)
        self.args['image'] = value

Implementation.register_class(Image)

class Text(Implementation):
    properties = dict(
        value=('text', HasProperties.prop_passthrough),
    )
    conprops = 'background foreground'.split()
    def create_widget(self):
        initial = self.args.get('text', '')
        kw = {}
        for name in self.conprops:
            if name in self.args:
                kw[name] = self.args[name]
        self.tk_widget = tk.Entry(self.parent.tk_widget, **kw)
        if initial:
            self.tk_widget.insert(0, initial)
    def set_value(self, value):
        self.tk_widget.delete(0, tk.END)
        self.tk_widget.insert(0, value)
    def get_value(self):
        return self.tk_widget.get()
Implementation.register_class(Text)

class Selection(Implementation):
    properties = dict(
        options=('options', HasProperties.prop_passthrough),
        value=('value', HasProperties.prop_passthrough),
    )
    conprops = 'background foreground'.split()
    def create_widget(self):
        self.tk_variable = tk.StringVar()
        options = self.args['options']
        value = self.args.get('value', options[0])
        self.tk_variable.set(value)

        args = (self.parent.tk_widget, self.tk_variable) + tuple(options)
        kw = {}
        for name in self.conprops:
            if name in self.args:
                kw[name] = self.args[name]
        self.tk_widget = tk.OptionMenu(*args, **kw)
Implementation.register_class(Selection)

class Button(Label):
    properties = dict(
        image=('image', HasProperties.prop_image),
        value=('text', HasProperties.prop_passthrough),
    )
    # TODO better name
    conprops = 'background foreground'.split()
    def create_widget(self):
        kw = {}
        for name in self.conprops:
            if name in self.args:
                kw[name] = self.args[name]
        kw['textvariable'] = self.tk_variable = tk.StringVar()
        self.tk_variable.set(self.args.get('text', ''))
        kw['command'] = functools.partial(
            self.spec.event_handlers.get('on_click'), self.spec)
        self.tk_widget = tk.Button(self.parent.tk_widget, **kw)
Implementation.register_class(Button)
Implementation.register_class(Button, 'Submit')
Implementation.register_class(Button, 'Cancel')

class Form(Implementation):
    def create_widget(self):
        self.tk_widget = tk.Frame(self.parent.tk_widget)

        for n, row in enumerate(self.spec.rows):
            if row.label is not None:
                l = Label(row.label, self)  # text=row.label.value, justify=tk.RIGHT)
                l.tk_widget.grid(row=n, column=0, sticky=tk.E)

            if row.widget is not None:
                # figure the name for this widget in the form
                w = Implementation.construct(row.widget, self)
                w.tk_widget.grid(row=n, column=1, sticky=tk.W)

            if row.help is not None:
                l = Label(row.help, self)  # text=row.help.value, justify=tk.LEFT)
                l.tk_widget.grid(row=n, column=2, sticky=tk.W)

        # switcheroo to put these buttons in a different parent
        old, self.tk_widget = self.tk_widget, tk.Frame(self.tk_widget)
        for button in self.spec.buttons:
            w = Implementation.construct(button, self)
            w.tk_widget.pack(side=tk.LEFT)
        self.tk_widget.grid(row=n+1, column=1, columnspan=2, sticky=tk.W)
        self.tk_widget = old

    def layout(self, child):
        pass

Implementation.register_class(Form, 'Form')



class CanvasObject(HasProperties):
    implementation_classes = {}
    @classmethod
    def register_class(cls, klass, name=None):
        if name is None:
            name = klass.__name__
        cls.implementation_classes[name] = klass

    @classmethod
    def construct(cls, spec, canvas):
        klass = cls.implementation_classes[spec.__class__.__name__]
        impl = klass(spec, canvas)

    def __init__(self, spec, canvas):
        self.canvas = canvas
        self.gui = canvas.gui
        self.spec = spec

        self.args = self.generate_properties(spec)
        self.x, self.y = position = self.args['x'], self.args['y']

        self.create_object(position, spec)

        if spec.event_handlers:
            self.canvas.event_handlers.append((spec, self.id,
                spec.event_handlers))
            # TODO self.handlers_required
        
        spec.implementation = self

    def set_value(self, value):
        raise NotImplementedError('set_value not implemented for %s'%
            self.tk_widget.type(self.id))

    def set_property(self, name, value):
        tname, value = self.translate_value(name, value)
        if name == 'value':
            self.set_value(value)
            return
        name = tname
        if name == 'x':
            pos = (value, self.y)
            self.canvas.tk_widget.coords(self.id, pos)
        elif name == 'y':
            pos = (self.x, value)
            self.canvas.tk_widget.coords(self.id, pos)
        else:
            self.canvas.tk_widget.itemconfig(self.id, **{name: value})

    def get_value(self):
        raise NotImplementedError('get_value not implemented for %s'%
            self.tk_widget.type(self.id))

    def get_property(self, name):
        if name == 'value':
            self.get_value()
        elif name == 'x':
            return self.canvas.tk_widget.coords(self.id)[0]
        elif name == 'y':
            return self.canvas.tk_widget.coords(self.id)[1]
        else:
            return self.canvas.tk_widget.itemcget(self.id, name)

    def destroy(self):
        self.canvas.delete(self)
        self.canvas = self.id = None

class CanvasImage(CanvasObject):
    properties = dict(
        value=('image', HasProperties.prop_image),
        x=('x', HasProperties.prop_int),
        y=('y', HasProperties.prop_int),
        anchor=('anchor', HasProperties.prop_anchor),
    )
    def create_object(self, position, spec):
        self.id = self.canvas.tk_widget.create_image(position,
            image=self.args['image'],
            anchor=self.args.get('anchor', tk.NW))
    def set_value(self, value):
        self.args['image'] = value
        return self.canvas.tk_widget.itemconfig(self.id, image=value)
    def get_value(self):
        return self.canvas.tk_widget.itemcget(self.id, 'image')
CanvasObject.register_class(CanvasImage, 'Image')

class CanvasLabel(CanvasObject):
    properties = dict(
        value=('text', HasProperties.prop_passthrough),
        x=('x', HasProperties.prop_int),
        y=('y', HasProperties.prop_int),
        anchor=('anchor', HasProperties.prop_anchor),
    )
    def create_object(self, position, spec):
        self.id = self.canvas.tk_widget.create_text(position,
            text=self.args.get('text', ''),
            anchor=self.args.get('anchor', tk.NW))
    def set_value(self, value):
        return self.canvas.tk_widget.itemconfig(self.id, text=value)
    def get_value(self):
        return self.canvas.tk_widget.itemcget(self.id, 'text')
CanvasObject.register_class(CanvasLabel, 'Label')

class Canvas(Implementation):
    properties = dict(
        width=('width', HasProperties.prop_passthrough),
        height=('height', HasProperties.prop_passthrough),
        background=('background', HasProperties.prop_color_or_image),
    )
    def create_widget(self):
        self.event_handlers = []

        # background image?
        image = self.args.get('background')
        if isinstance(image, ImageTk.PhotoImage):
            if 'width' not in self.args:
                self.args['width'] = image.width()
            if 'height' not in self.args:
                self.args['height'] = image.height()
        elif image is not None:
            kw['background'] = image
            image = None

        # basic args
        kw = {}
        for name in 'width height'.split():
            if name in self.args:
                kw[name] = self.args[name]

        # create the widget
        self.tk_widget = tk.Canvas(self.parent.tk_widget, **kw)

        # add bg image?
        if image is not None:
            self.bg_image_id = self.tk_widget.create_image((0,0),
                image=image, anchor=tk.NW)

        for child in self.spec.children:
            CanvasObject.construct(child, self)

        # TODO if mouse required
        self.tk_widget.bind("<Button-1>", self.on_mouse)

    def on_mouse(self, event):
        for w, id, handlers in self.event_handlers:
            if 'on_mouse' not in handlers: continue
            x1, y1, x2, y2 = self.tk_widget.bbox(id)
            if event.x < x1 or event.x > x2: continue
            if event.y < y1 or event.y > y2: continue
            MouseEvent(w, handlers['on_mouse'])(event)
            break

    def delete(self, child):
        self.event_handlers = [h
            for h in self.event_handlers
                if h[1] != child.id
        ]
        self.tk_widget.delete(child.id)
    
    def add_child(self, spec):
        return CanvasObject.construct(spec, self)

Implementation.register_class(Canvas)

class MouseEvent(object):
    def __init__(self, spec, action, left=False, middle=False, right=False):
        self.spec = spec
        self.left = left
        self.middle = middle
        self.right = right
        self.action = action
        
    def __call__(self, event):
        self.x , self.y = event.x, event.y
        self.action(self.spec, self)

